package ru.dbotthepony.kstarbound.client.gl.shader

import com.google.common.collect.ImmutableList
import org.lwjgl.opengl.GL20.GL_FRAGMENT_SHADER
import org.lwjgl.opengl.GL20.GL_VERTEX_SHADER
import ru.dbotthepony.kstarbound.client.StarboundClient
import ru.dbotthepony.kstarbound.client.gl.vertex.GeometryType
import ru.dbotthepony.kstarbound.client.gl.vertex.StreamVertexBuilder
import ru.dbotthepony.kstarbound.client.gl.vertex.VertexAttributes
import ru.dbotthepony.kstarbound.client.render.FONT_VERTEX_FORMAT
import ru.dbotthepony.kvector.api.IStruct4f
import ru.dbotthepony.kvector.arrays.Matrix3f
import ru.dbotthepony.kvector.vector.RGBAColor

private fun shaders(name: String): List<GLShader> {
	val client = StarboundClient.current()
	return listOf(client.internalVertex("shaders/$name.vsh"), client.internalFragment("shaders/$name.fsh"))
}

private fun shaders1(name: String): List<GLShader> {
	val client = StarboundClient.current()
	return listOf(client.internalVertex("shaders/vertex/$name.vsh"), client.internalFragment("shaders/fragment/$name.fsh"))
}

private fun fontShaders(): List<GLShader> {
	val client = StarboundClient.current()
	var read = StarboundClient.readInternal("shaders/fragment/font.fsh")

	val textures = ArrayList<String>()
	val cases = ArrayList<String>()
	val cases2 = ArrayList<String>()

	for (i in 0 until client.maxUserTextureBlocks) {
		cases.add("\tvec4 colorResult$i = colorMultiplier * texture(texture$i, uvOut).r;")
	}

	for (i in 0 until client.maxUserTextureBlocks) {
		textures.add("layout (binding=$i) uniform sampler2D texture$i;")
		cases2.add("if (textureIndexOut < ${i + 0.1f}) { colorResult = colorResult$i; }")
	}

	cases2.add(" { colorResult = colorMultiplier; }")

	read = read.replace("uniform sampler2D texture0;", textures.joinToString("\n")).replace("\tcolorResult = colorMultiplier * texture(texture0, uvOut).r;", cases.joinToString("\n") + cases2.joinToString("\n\telse "))

	val fragment = GLShader(read, GL_FRAGMENT_SHADER)
	val vertex = GLShader(StarboundClient.readInternal("shaders/vertex/font.vsh"), GL_VERTEX_SHADER)

	return listOf(fragment, vertex)
}

class FontProgram : GLShaderProgram(shaders1("font"), FONT_VERTEX_FORMAT), GLShaderProgram.Regular {
	override var viewMatrix: Matrix3f by F3x3Uniform("viewMatrix")
	override var worldMatrix: Matrix3f by F3x3Uniform("worldMatrix")
	override var modelMatrix: Matrix3f by F3x3Uniform("modelMatrix", true)
	override var colorMultiplier: IStruct4f by F4Uniform("colorMultiplier")

	var texture by IUniform("texture0")

	init {
		viewMatrix = Matrix3f.identity()
		modelMatrix = Matrix3f.identity()
		colorMultiplier = RGBAColor.WHITE
	}
}

class GLLiquidProgram : GLShaderProgram(shaders("liquid"), VertexAttributes.POSITION) {
	var baselineColor by F4Uniform("baselineColor")
	var transform by F3x3Uniform("transform")
	val builder = StreamVertexBuilder(attributes, GeometryType.QUADS)
}

class GLPrograms {
	val position = UberShader.Builder().build()
	val positionTexture = UberShader.Builder().withTexture().build()
	val positionTextureLightmap = UberShader.Builder().withTexture().withLightMap().build()
	val positionColor = UberShader.Builder().withColor().build()
	val tile = UberShader.Builder().withTexture().withHueShift().withLightMap().build()
	val font = FontProgram()
	val liquid = GLLiquidProgram()
}
